# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
import string
from time import time, sleep

from typing import List, Iterable, Tuple, Union

import allure
import pytest
from _pytest.fixtures import FixtureFunctionMarker, _FixtureFunction
from _pytest.mark import MarkDecorator
from adcm_client.base import ObjectNotFound
from adcm_client.objects import Host, Task, Cluster


def remove_host(host: Host) -> Task:
    """
    Run action remove for host
    """
    return host.action(name="remove").run()


def get_or_add_service(cluster: Cluster, service_name: str):
    """
    Add service if it wasn't added before and return it
    """
    try:
        return cluster.service(name=service_name)
    except ObjectNotFound:
        return cluster.service_add(name=service_name)


def random_string(strlen: int = 10) -> str:
    """
    Generate random string

    >>> type(random_string())
    <class 'str'>
    >>> all([True if char in string.ascii_letters else False for char in random_string()])
    True
    >>> len(random_string())
    10
    >>> len(random_string(strlen=0))
    0
    >>> len(random_string(strlen=1))
    1
    """
    return "".join([random.choice(string.ascii_letters) for _ in range(strlen)])


def random_special_chars(strlen: int = 3) -> str:
    """
    Generate special chars

    >>> type(random_special_chars())
    <class 'str'>
    >>> all([True if char in string.punctuation else False for char in random_special_chars()])
    True
    >>> len(random_special_chars())
    3
    >>> len(random_special_chars(strlen=0))
    0
    >>> len(random_special_chars(strlen=1))
    1
    """
    return "".join([random.choice(string.punctuation) for _ in range(strlen)])


def random_string_list(num: int = 10) -> List[str]:
    """
    Generate list of random strings

    >>> type(random_string_list())
    <class 'list'>
    >>> all([True if type(obj) is str else False for obj in random_string_list()])
    True
    >>> len(random_string_list())
    10
    >>> len(random_string_list(num=0))
    0
    >>> len(random_string_list(num=1))
    1
    """
    return [random_string() for _ in range(num)]


def ordered_dict_to_dict(value: dict) -> dict:
    """
    Convert OrderedDict to dict

    :param value: some OrderedDict object

    >>> ordered_dict_to_dict({})
    {}
    >>> ordered_dict_to_dict({'key': 'value'})
    {'key': 'value'}
    >>> ordered_dict_to_dict({'key': {'nested_key': 'nested_value'}})
    {'key': {'nested_key': 'nested_value'}}
    >>> from collections import OrderedDict
    >>> d = {'banana': 3, 'apple': 4, 'pear': 1, 'orange': 2}
    >>> ordered_dict_to_dict(OrderedDict(d.items()))
    {'banana': 3, 'apple': 4, 'pear': 1, 'orange': 2}
    >>> d = {'banana': 3, 'apple': 4, 'pear': 1, 'orange': {'some_key': 'some_value'}}
    >>> ordered_dict_to_dict(OrderedDict(d.items()))
    {'banana': 3, 'apple': 4, 'pear': 1, 'orange': {'some_key': 'some_value'}}
    >>> d = {'banana': 3, 'apple': 4, 'pear': 1, 'orange': 2}
    >>> ordered_dict_to_dict(OrderedDict(sorted(d.items(), key=lambda t: t[0])))
    {'apple': 4, 'banana': 3, 'orange': 2, 'pear': 1}
    """
    for k, v in value.items():
        if isinstance(v, dict):
            value[k] = ordered_dict_to_dict(v)
    return dict(value)


def deep_merge(first: dict, second: dict) -> dict:
    """
    Dict deep merge function
    Merge recursive second dict in first and return it

    >>> deep_merge({},{})
    {}
    >>> deep_merge({'key': 'value'}, {'key': 'value'})
    {'key': 'value'}
    >>> deep_merge({'key': 'value'}, {'key': 'new_value', 'second_key': 'second_value'})
    {'key': 'new_value', 'second_key': 'second_value'}
    >>> deep_merge(
    ...     {'key': 'value', 'dict_key': {'nested_key': 'nested_value'}},
    ...     {
    ...         'key': 'value',
    ...         'dict_key': {'nested_key': 'new_nested_value', 'nested_dict_key': {}},
    ...         'flat_key': 'flat_value'
    ...     }
    ... )
    {'key': 'value', 'dict_key': {'nested_key': 'new_nested_value', 'nested_dict_key': {}}, 'flat_key': 'flat_value'}
    """
    for key in second:
        if key in first:
            if isinstance(first[key], dict) and isinstance(second[key], dict):
                deep_merge(first[key], second[key])
            else:
                first[key] = second[key]
        else:
            first[key] = second[key]
    return first


def check_mutually_exclusive(options, *opts) -> bool:
    """
    Checks if there are mutually exclusive options

    >>> obj = lambda: None
    >>> obj.attrib1 = 1
    >>> obj.attrib2 = 2
    >>> check_mutually_exclusive(obj, *["attrib1"])
    False
    >>> obj = lambda: None
    >>> obj.attrib1 = 1
    >>> obj.attrib2 = 2
    >>> obj.attrib3 = 3
    >>> check_mutually_exclusive(obj, *["attrib1", "attrib2"])
    True
    """
    count = 0
    for opt in opts:
        if getattr(options, opt):
            count += 1
    return count > 1


def get_subdirs_iter(filename: str, *subdirs) -> Iterable[str]:
    """
    Get iterable subdirs

    >>> type(get_subdirs_iter(__file__))
    <class 'generator'>
    >>> type(get_subdirs_iter(__file__, "subdir1", "subdir2"))
    <class 'generator'>
    """

    datadir = get_data_dir(filename)
    dirname = os.path.join(datadir, *subdirs)
    for b in os.listdir(dirname):
        full = os.path.join(dirname, b)
        if os.path.isdir(full):
            yield full


def get_data_dir(filename: str, *subdirs) -> str:
    """That function returns a name of data dir for test.

    It's nice for every test to store test data in <module_name>_data dir.
    That function returns path to that dir.
    Optionally it joins data dir with list of subdirs. That allow to do
    one call to build a path

    Example:

    get_data_dir(__file__, "bundles", "cluster")

    Return will be:
    "<filename>/bundles/cluster"

    NOTE: That function doesn't check existing of a path.

    >>> get_data_dir(__file__).endswith("_data")
    True
    >>> get_data_dir(__file__, "subdir1", "subdir2").endswith("_data/subdir1/subdir2")
    True
    >>> __file__ == f"{get_data_dir(__file__).split('_data')[0]}.py"
    True
    """
    filename = filename[:-3]  # Strip .py from name
    filename = filename + "_data"
    filename = os.path.join(filename, *subdirs)
    return filename


def get_data_subdirs_as_parameters(
    filename: str, *subdirs
) -> Tuple[List[str], List[str]]:
    """That function returns subdirs of <filename>_data in parametrize form

    It's really useful to iterate over subdirs with pytest.mark.parametrize.
    It returns two list:
    - list with pathes
    - list with names

    You could use it in code like that
    cases, ids = get_data_subdirs_as_parameters(__file__)
    @pytest.mark.parametrize("subdir", cases, ids=ids)

    Besides, it is possible to go deeply with the os.path.join, which is called
    against any argument from second
    cases, ids = get_data_subdirs_as_parameters(__file__, "subdirname", "subsubdirname")
    @pytest.mark.parametrize("subdir", cases, ids=ids)

    >>> @contextlib.contextmanager
    ... def _mk_dir(tmp_dir):
    ...     os.mkdir(tmp_dir)
    ...     yield tmp_dir
    ...     shutil.rmtree(tmp_dir)
    >>> test_dir = f"{__file__.strip('.py')}_data"
    >>> with _mk_dir(test_dir):
    ...     get_data_subdirs_as_parameters(__file__)
    ([], [])
    >>> dirs = [f"{test_dir}/level1", f"{test_dir}/level1/level2_0", f"{test_dir}/level1/level2_1"]
    >>> with _mk_dir(test_dir), _mk_dir(dirs[0]), _mk_dir(dirs[1]), _mk_dir(dirs[2]):
    ...         get_data_subdirs_as_parameters(__file__, "level1") == ( dirs[1:], ['level2_0', "level2_1"])
    True
    """
    datadir = get_data_dir(filename)
    dirname = os.path.join(datadir, *subdirs)
    paths = []
    ids = []
    for b in sorted(os.listdir(dirname)):
        paths.append("{}/{}".format(dirname, b))
        ids.append(b)
    return paths, ids


def parametrize_by_data_subdirs(filename: str, *path) -> MarkDecorator:
    """This is a decorator useful to combine functionality of
    get_data_subdirs_as_parameters and pytest.mark.parametrize.

    Example:

    @parametrize_by_data_subdirs(__file__)
    def test_over_dirs(path):
        pass

    In that case function find directory <filename>_data and parametrize(iterate)
    test case over subdirs.

    Sometimes you need to go deeper than in _data folder:

    @parametrize_by_data_subdirs(__file__, 'subdir_leve1', 'subdir_level2'):
    def test_over_dirs(path):
        pass

    In that case test will iterate over <filename>_data/subdir_leve1/subdir_level2/* dirs.

    >>> @contextlib.contextmanager
    ... def _mk_dir(tmp_dir):
    ...     os.mkdir(tmp_dir)
    ...     yield tmp_dir
    ...     shutil.rmtree(tmp_dir)
    >>> test_dir = f"{__file__.strip('.py')}_data"
    >>> dirs = [f"{test_dir}/level1", f"{test_dir}/level1/level2_0", f"{test_dir}/level1/level2_1"]
    >>> with _mk_dir(test_dir), _mk_dir(dirs[0]), _mk_dir(dirs[1]), _mk_dir(dirs[2]):
    ...         mark_decorator = parametrize_by_data_subdirs(__file__, "level1")
    >>> mark_decorator.kwargs
    {'ids': ['level2_0', 'level2_1']}
    >>> mark_decorator.args == ('path', dirs[1:])
    True
    """
    cases, ids = get_data_subdirs_as_parameters(filename, *path)
    return pytest.mark.parametrize("path", cases, ids=ids)


def fixture_parametrized_by_data_subdirs(
    filename: str, *path, scope="function"
) -> Union[FixtureFunctionMarker, _FixtureFunction]:
    """That is a combination of parametrized fixture and get_data_subdirs_as parameters

    That is useful when you want to parametrize fixture over subdirs.

    Example:

    @fixture_parametrized_by_data_subdirs(__file__, 'cluster_and_service', scope='module')
    def cluster(sdk_client_ms: ADCMClient, request):
        assert request.param is not None

    Note:

    Due to strange pytest logic you can't parametrize a fixture with right way
    (over regular pytest.mark.parametrize decorator). That is why we have strange
    access to information over request.param.

    >>> @contextlib.contextmanager
    ... def _mk_dir(tmp_dir):
    ...     os.mkdir(tmp_dir)
    ...     yield tmp_dir
    ...     shutil.rmtree(tmp_dir)
    >>> test_dir = f"{__file__.strip('.py')}_data"
    >>> with _mk_dir(test_dir):
    ...     fixture_function_marker = fixture_parametrized_by_data_subdirs(__file__)
    >>> fixture_function_marker.scope
    'function'
    >>> fixture_function_marker.ids
    ()
    >>> fixture_function_marker.params
    ()
    >>> dirs = [f"{test_dir}/level1", f"{test_dir}/level1/level2_0", f"{test_dir}/level1/level2_1"]
    >>> with _mk_dir(test_dir), _mk_dir(dirs[0]), _mk_dir(dirs[1]), _mk_dir(dirs[2]):
    ...         fixture_function_marker = fixture_parametrized_by_data_subdirs(
    ...             __file__, "level1", scope='module'
    ...         )
    >>> fixture_function_marker.scope
    'module'
    >>> fixture_function_marker.ids
    ('level2_0', 'level2_1')
    >>> fixture_function_marker.params == tuple(dirs[1:])
    True
    """
    cases, ids = get_data_subdirs_as_parameters(filename, *path)
    return pytest.fixture(scope=scope, params=cases, ids=ids)


def wait_until_step_succeeds(
    func, timeout: Union[int, float] = 300, period: Union[int, float] = 10, **kwargs
):
    """
    Repeat `func` with `kwargs` until successful
    >>> states = iter([False, False, True])
    >>> def step():
    ...     assert next(states), "Still failed"

    >>> wait_until_step_succeeds(step, timeout=0.5, period=0.1)
    >>> states = iter([False, False, True])
    >>> wait_until_step_succeeds(step, timeout=0.2, period=0.1)
    Traceback (most recent call last):
    ...
    AssertionError: Step "step" failed after retrying 0.2 seconds. The last error was: Still failed
    """
    with allure.step(f'Wait until "{func.__name__}" succeeds'):
        start = time()
        last_error = None
        while time() - start < timeout:
            try:
                func(**kwargs)
            except AssertionError as e:
                last_error = e
                sleep(period)
                continue
            break
        else:
            raise AssertionError(
                f'Step "{func.__name__}" failed after retrying {timeout} seconds. '
                f"The last error was: {last_error}"
            )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Optional

import allure
import docker
import ifaddr
import pytest
import socket

from adcm_client.base import Paging, WaitTimeout
from adcm_client.objects import ADCMClient
from allure_commons.reporter import AllureReporter
from allure_commons.utils import uuid4
from allure_pytest.listener import AllureListener
from requests.exceptions import ReadTimeout as DockerReadTimeout
from retry.api import retry_call
from deprecated import deprecated

from .docker_utils import (
    ADCM,
    DockerWrapper,
    is_docker,
    gather_adcm_data_from_container,
    get_initialized_adcm_image,
    split_tag,
)
from .utils import check_mutually_exclusive, remove_host

__all__ = [
    "image",
    "cmd_opts",
    "adcm",
    "client",
    "adcm_fs",
    "adcm_ss",
    "adcm_ms",
    "sdk_client_ms",
    "sdk_client_fs",
    "sdk_client_ss",
]


# pylint: disable=W0621
@allure.title("ADCM Image")
@pytest.fixture(scope="session")
def image(request, cmd_opts):
    """That fixture creates ADCM container, waits until
    a database becomes initialised and stores that as images
    with random tag and name local/adcminit

    That can be useful to use that fixture to make ADCM's
    container startup time shorter.

    Operates with cmd-opts:
     '--staticimage INIT_IMAGE'
     '--adcm-image IMAGE'
     '--remote-docker HOST:PORT'
     '--dontstop'
     '--nopull'

    Fixture returns list:
    repo, tag
    """

    mutually_exclusive_opts = [
        ["adcm_image", "adcm_images", "adcm_min_version"],
        ["staticimage", "adcm_images", "adcm_min_version"],
    ]
    # if more than one option that defines image params is used raise exception
    # pytest don't allow more convenient mechanisms to add mutually exclusive options.
    for opt_sets in mutually_exclusive_opts:
        if check_mutually_exclusive(cmd_opts, *opt_sets):
            raise Exception(
                "wrong using of import parameters %s are mutually exclusive"
                % ", ".join(opt_sets)
            )

    pull = not cmd_opts.nopull
    if cmd_opts.remote_docker:
        dc = docker.DockerClient(
            base_url=f"tcp://{cmd_opts.remote_docker}", timeout=120
        )
    else:
        dc = docker.from_env(timeout=120)

    params = dict()
    if cmd_opts.staticimage:
        params["repo"], params["tag"] = split_tag(cmd_opts.staticimage)
    # if image fixture was indirectly parametrized
    # use 'adcm_repo' and 'adcm_tag' from parametrisation
    if hasattr(request, "param"):
        params["adcm_repo"], params["adcm_tag"] = request.param
    # if there is no parametrization check if adcm_image option is passed
    elif cmd_opts.adcm_image:
        params["adcm_repo"], params["adcm_tag"] = split_tag(cmd_opts.adcm_image)

    if not (cmd_opts.dontstop or cmd_opts.staticimage):

        def fin():
            if init_image:
                image_name = "{}:{}".format(*init_image.values())
                for container in dc.containers.list(filters=dict(ancestor=image_name)):
                    try:
                        container.wait(condition="removed", timeout=30)
                    except ConnectionError:
                        # https://github.com/docker/docker-py/issues/1966 workaround
                        pass
                retry_call(
                    dc.images.remove,
                    fargs=[image_name],
                    fkwargs={"force": True},
                    tries=5,
                )

        # Set None for init image to avoid errors in finalizer
        # when get_initialized_adcm_image() fails
        init_image = None
        request.addfinalizer(fin)

    init_image = get_initialized_adcm_image(pull=pull, dc=dc, **params)

    return init_image["repo"], init_image["tag"]


def _adcm(image, cmd_opts, request) -> ADCM:
    repo, tag = image
    if cmd_opts.remote_docker:
        dw = DockerWrapper(base_url=f"tcp://{cmd_opts.remote_docker}")
        ip = cmd_opts.remote_docker.split(":")[0]
    else:
        dw = DockerWrapper()
        ip = (
            _get_connection_ip(cmd_opts.remote_executor_host)
            if cmd_opts.remote_executor_host
            else None
        )
        if ip and is_docker():
            if _get_if_type(ip) == "0":
                raise EnvironmentError(
                    "You are using network interface with 'bridge' "
                    "type while running inside container."
                    "There is no obvious way to get external ip in this case."
                    "Try running container with pytest with --net=host option"
                )
    adcm = dw.run_adcm(
        image=repo,
        tag=tag,
        pull=False,
        ip=ip,
        labels={"pytest_node_id": request.node.nodeid},
    )

    def fin():
        if not request.config.option.dontstop:
            gather = True
            try:
                if not request.node.rep_call.failed:
                    gather = False
            except AttributeError:
                # There is no rep_call attribute. Presumably test setup failed,
                # or fixture scope is not function. Will collect /adcm/data anyway
                pass
            if gather:
                with allure.step("Gather /adcm/data/ from ADCM container"):
                    file_name = f"ADCM Log {request.node.name}_{time.time()}"
                    reporter = _allure_reporter(request.config)
                    if reporter:
                        test_result = reporter.get_test(uuid=None)
                        with gather_adcm_data_from_container(adcm) as data:
                            reporter.attach_data(
                                uuid=uuid4(),
                                body=data,
                                name="{}.tgz".format(file_name),
                                extension="tgz",
                                parent_uuid=test_result.uuid,
                            )
                    else:
                        with gather_adcm_data_from_container(adcm) as data:
                            allure.attach(
                                body=data,
                                name="{}.tgz".format(file_name),
                                extension="tgz",
                            )

            _remove_hosts(adcm)

            try:
                adcm.container.kill()
            except DockerReadTimeout:
                pass

    request.addfinalizer(fin)

    adcm.api.auth(username="admin", password="admin")

    return adcm


def _allure_reporter(config) -> Optional[AllureReporter]:
    """Get Allure Reporter from pytest plugins"""
    listener: AllureListener = next(
        filter(
            lambda plugin: (isinstance(plugin, AllureListener)),
            dict(config.pluginmanager.list_name_plugin()).values(),
        ),
        None,
    )
    return listener.allure_logger if listener else None


def _get_connection_ip(remote_host: str):
    """
    Try to open connection to remote and get ip address of the interface used.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect((remote_host, 1))
    ip = s.getsockname()[0]
    s.close()
    return ip


def _get_if_name_by_ip(if_ip):
    """Get interface name by interface IP"""
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            if ip.ip == if_ip:
                return adapter.name
    raise ValueError(f"IP {if_ip} does not match any network interface!")


def _get_if_type(if_ip):
    """
    Get interface type from /sys/class/net/{if_name}/type
    """
    if_name = _get_if_name_by_ip(if_ip)
    with open(f"/sys/class/net/{if_name}/type", "r") as f:
        return f.readline().strip()


def _remove_hosts(adcm: ADCM):
    client = ADCMClient(api=adcm.api)
    for cluster in Paging(client.cluster_list):
        cluster.delete()
    jobs = list()
    for host in Paging(client.host_list):
        if host.state != "removed" and "remove" in list(
            map(lambda x: getattr(x, "name"), host.action_list())
        ):
            jobs.append(remove_host(host))
    if jobs:
        for job in jobs:
            try:
                job.wait(timeout=60)
            # I case when host were not removed in requested timeout
            # we don't want it to affect test result
            except WaitTimeout:
                pass


# Legacy fixture should not be used in new tests!
@deprecated
@pytest.fixture(scope="module")
def adcm(image, cmd_opts, request):
    """Legacy ADCM object fixture. Do not use in new tests"""
    return _adcm(image, cmd_opts, request)


# Legacy fixture should not be used in new tests!
@deprecated
@pytest.fixture(scope="module")
def client(adcm):
    """Legacy api object fixture. Do not use in new tests"""
    return adcm.api.objects


##################################################
#                  S D K
##################################################
@allure.title("[MS] ADCM Container")
@pytest.fixture(scope="module")
def adcm_ms(image, cmd_opts, request) -> ADCM:
    """Runs adcm container from the previously initialized image.
    Operates '--dontstop' option.
    Returns authorized instance of ADCM object
    """
    return _adcm(image, cmd_opts, request)


@allure.title("[FS] ADCM Container")
@pytest.fixture(scope="function")
def adcm_fs(image, cmd_opts, request) -> ADCM:
    """Runs adcm container from the previously initialized image.
    Operates '--dontstop' option.
    Returns authorized instance of ADCM object
    """
    return _adcm(image, cmd_opts, request)


@allure.title("[SS] ADCM Container")
@pytest.fixture(scope="session")
def adcm_ss(image, cmd_opts, request) -> ADCM:
    """Runs adcm container from the previously initialized image.
    Operates '--dontstop' option.
    Returns authorized instance of ADCM object
    """
    return _adcm(image, cmd_opts, request)


@allure.title("[MS] ADCM Client")
@pytest.fixture(scope="module")
def sdk_client_ms(adcm_ms: ADCM) -> ADCMClient:
    """Returns ADCMClient object from adcm_client"""
    return ADCMClient(api=adcm_ms.api)


@allure.title("[FS] ADCM Client")
@pytest.fixture(scope="function")
def sdk_client_fs(adcm_fs: ADCM) -> ADCMClient:
    """Returns ADCMClient object from adcm_client"""
    return ADCMClient(api=adcm_fs.api)


@allure.title("[SS] ADCM Client")
@pytest.fixture(scope="session")
def sdk_client_ss(adcm_ss: ADCM) -> ADCMClient:
    """Returns ADCMClient object from adcm_client"""
    return ADCMClient(api=adcm_ss.api)


@allure.title("Pytest options")
@pytest.fixture(scope="session")
def cmd_opts(request):
    """Returns pytest request options object"""
    return request.config.option



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common test steps with actions
"""
from contextlib import contextmanager
from difflib import get_close_matches
from typing import Union
from coreapi.exceptions import ErrorMessage

import allure
from adcm_client.base import ObjectNotFound
from adcm_client.objects import Cluster, Service, Host, Task, Component, Provider

from .asserts import assert_action_result


def _get_error_text_from_task_logs(task: Task):
    error_text = ""
    for job in task.job_list():
        try:
            for log in job.log_list():
                if log.type == "stdout":
                    error_text += _extract_error_from_ansible_log(log.content)
        except ErrorMessage as log_exception:
            # Multijobs has no logs for parent Job
            # pylint: disable=protected-access
            if log_exception.error._data["code"] == "LOG_NOT_FOUND":
                pass
            else:
                raise log_exception
    return error_text


def _extract_error_from_ansible_log(log: str):
    """
    Extract ansible error message from ansible log

    >>> _extract_error_from_ansible_log("fatal: Some ERROR\\nTASK [api : something] **********")
    'fatal: Some ERROR\\n'
    >>> _extract_error_from_ansible_log(
    ...     "TASK [api : something] **********\\nok: [adcm-cluster-adb-gw0-e330benhwqir]\\n msg: All assertions passed"
    ... )
    ''
    >>> _extract_error_from_ansible_log(
    ...     "TASK [conf]**********\\nfatal: Some \\n multiline\\nERROR\\nNO MORE HOSTS LEFT *********"
    ... )
    'fatal: Some \\n multiline\\nERROR\\n'
    """
    err_start = log.find("fatal:")
    if err_start > -1:
        task_marker = log.find("******", err_start)
        err_end = log.rfind("\n", 0, task_marker) + 1
        return log[err_start:err_end]
    return ""


def _run_action_and_assert_result(
    obj: Union[Cluster, Service, Host, Component, Provider],
    action_name: str,
    expected_status="success",
    **kwargs,
):
    """
    Run action and assert that status equals to 'status' argument
    """
    obj_name = obj.name if not isinstance(obj, Host) else obj.fqdn
    with allure.step(
        f"Perform action '{action_name}' for {obj.__class__.__name__} '{obj_name}'"
    ), _suggest_action_if_not_exists(obj, action_name):
        task = obj.action(name=action_name).run(**kwargs)
        result = task.wait()
        ansible_error = ""
        if result != expected_status and expected_status == "success":
            ansible_error = _get_error_text_from_task_logs(task)
        assert_action_result(
            name=action_name,
            result=result,
            status=expected_status,
            additional_message=ansible_error,
        )


def run_cluster_action_and_assert_result(
    cluster: Cluster, action: str, status="success", **kwargs
):
    """
    Run cluster action and assert that status equals to 'status' argument
    """
    _run_action_and_assert_result(
        obj=cluster, action_name=action, expected_status=status, **kwargs
    )


def run_service_action_and_assert_result(
    service: Service, action: str, status="success", **kwargs
):
    """
    Run service action and assert that status equals to 'status' argument
    """
    _run_action_and_assert_result(
        obj=service, action_name=action, expected_status=status, **kwargs
    )


def run_component_action_and_assert_result(
    component: Component, action: str, status="success", **kwargs
):
    """
    Run component action and assert that status equals to 'status' argument
    """
    _run_action_and_assert_result(
        obj=component, action_name=action, expected_status=status, **kwargs
    )


def run_host_action_and_assert_result(
    host: Host, action: str, status="success", **kwargs
):
    """
    Run host action and assert that status equals to 'status' argument
    """
    _run_action_and_assert_result(
        obj=host, action_name=action, expected_status=status, **kwargs
    )


def run_provider_action_and_assert_result(
    provider: Provider, action: str, status="success", **kwargs
):
    """
    Run provider action and assert that status equals to 'status' argument
    """
    _run_action_and_assert_result(
        obj=provider, action_name=action, expected_status=status, **kwargs
    )


@contextmanager
def _suggest_action_if_not_exists(
    obj: Union[Cluster, Service, Host, Component], action
):
    """
    Contextmanager that suggest action if for some reason the required action was not found
    """
    try:
        yield
    except ObjectNotFound as e:
        all_actions = [a.name for a in obj.action_list()]
        suggest_actions = ", ".join(get_close_matches(action, all_actions))
        if suggest_actions:
            raise ObjectNotFound(
                f"No such action {action}. Did you mean: {suggest_actions}?"
            ) from e
        all_actions = ", ".join(all_actions)
        raise ObjectNotFound(
            f"No such action {action}. Possible actions: {all_actions}."
        ) from e


def wait_for_task_and_assert_result(task: Task, status: str):
    """
    Wait for a task to be completed and assert status to be equal to 'status' argument
    """
    # Need to add Action name instead of task_id when links will be provided in ADCM SDK
    with allure.step("Wait for a task to be completed"):
        result = task.wait()
        assert_action_result(result=result, status=status)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest


class Layer:
    """
    Decorators of allure layers
    Example:

        @Layer.UI
        def test_ui():
            pass


        @Layer.API
        def test_api():
            pass
    """

    UI = pytest.mark.allure_label("UI", label_type="layer")
    API = pytest.mark.allure_label("API", label_type="layer")
    Unit = pytest.mark.allure_label("Unit", label_type="layer")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common test asserts
"""

import allure

from adcm_client.objects import BaseAPIObject


def assert_state(obj: BaseAPIObject, state):
    """
    Asserts object state to be equal to 'state' argument

    >>> some_obj = lambda: None
    >>> some_obj.reread = lambda: None
    >>> some_obj.state = "installed"
    >>> assert_state(some_obj, "installed") is None
    True
    >>> assert_state(some_obj, "started") # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    AssertionError:
    """
    obj.reread()
    if hasattr(obj, "name"):
        name = obj.name
    elif hasattr(obj, "fqdn"):
        name = obj.fqdn
    else:
        name = obj.__repr__()
    with allure.step(
        f"Assert state of {obj.__class__.__name__} '{name}' to be equal to '{state}'"
    ):
        assert obj.state == state, (
            f"Object '{name}' have unexpected state - '{obj.state}'. "
            f"Expected - '{state}'"
        )


@allure.step("Assert action result to be equal to {status}")
def assert_action_result(result: str, status: str, name="", additional_message=""):
    """
    Asserts action result to be equal to 'status' argument

    >>> assert_action_result("200", "200") is None
    True
    >>> assert_action_result("200", "400")
    Traceback (most recent call last):
    ...
    AssertionError: Action  finished execution with unexpected result - '200'. Expected - '400'
    >>> assert_action_result("200", "400", "some_action")
    Traceback (most recent call last):
    ...
    AssertionError: Action some_action finished execution with unexpected result - '200'. Expected - '400'
    >>> assert_action_result("200", "500", additional_message="My custom message")
    Traceback (most recent call last):
    ...
    AssertionError: Action  finished execution with unexpected result - '200'. Expected - '500'
    My custom message
    """
    message = (
        f"Action {name} finished execution with unexpected result - '{result}'. "
        f"Expected - '{status}'"
    )
    if additional_message:
        message += f"\n{additional_message}"
    assert result == status, message



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name, C0103, E0401
import io
import os
import re
import random
import socket
from contextlib import contextmanager
from gzip import compress

import allure
import docker
import tarfile

from docker.errors import APIError, ImageNotFound
from adcm_client.util.wait import wait_for_url
from adcm_client.wrappers.api import ADCMApiWrapper

from .utils import random_string

MIN_DOCKER_PORT = 8000
MAX_DOCKER_PORT = 9000
DEFAULT_IP = "127.0.0.1"
CONTAINER_START_RETRY_COUNT = 20


class UnableToBind(Exception):
    pass


class RetryCountExceeded(Exception):
    pass


def _port_is_free(ip, port) -> bool:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((ip, port))
    if result == 0:
        return False
    return True


def _find_random_port(ip) -> int:
    for _ in range(0, 20):
        port = random.randint(MIN_DOCKER_PORT, MAX_DOCKER_PORT)
        if _port_is_free(ip, port):
            return port
    raise UnableToBind("There is no free port for Docker after 20 tries.")


def is_docker() -> bool:
    """
    Look into cgroup to detect if we are in container
    """
    path = "/proc/self/cgroup"
    try:
        with open(path) as f:
            for line in f:
                if re.match(r"\d+:[\w=]+:/docker(-[ce]e)?/\w+", line):
                    return True
    except FileNotFoundError:
        pass
    return False


@contextmanager
def gather_adcm_data_from_container(adcm: "ADCM"):
    """
    Get /adcm/data/ form ADCM docker container
    :return: compressed file stream
    """
    bits, _ = adcm.container.get_archive("/adcm/data/")

    with io.BytesIO() as stream:
        for chunk in bits:
            stream.write(chunk)
        stream.seek(0)
        yield compress(stream.getvalue())


def get_file_from_container(instance, path, filename):
    """
    Get file from docker container and return file object

    Args:
        instance: ADCM instance
        path (str): path to file in container
        filename (str): filename in path

    Returns:
        (file object): The extracted file from tar archive from docker container

    """

    stream = instance.container.get_archive(path + filename)[0]
    file_obj = io.BytesIO()
    for i in stream:
        file_obj.write(i)
    file_obj.seek(0)
    tar = tarfile.open(mode="r", fileobj=file_obj)
    return tar.extractfile(filename)


@allure.step("Prepare initialized ADCM image")
def get_initialized_adcm_image(
    repo="local/adcminit",
    tag=None,
    adcm_repo=None,
    adcm_tag=None,
    pull=True,
    dc=None,
) -> dict:
    """
    We consider that if we know tag, staticimage option is used,
    container is already initialized. In case if option is used but image is absent
    we will create an image with such tag for further use.
    If we don't know tag image must be initialized, tag will be randomly generated.
    """
    if not dc:
        dc = docker.from_env(timeout=120)

    if tag and image_exists(repo, tag, dc):
        return {"repo": repo, "tag": tag}
    else:
        if not tag:
            tag = random_string()
        return init_adcm(repo, tag, adcm_repo, adcm_tag, pull, dc)


def init_adcm(repo, tag, adcm_repo, adcm_tag, pull, dc=None):
    dw = DockerWrapper(dc=dc)
    # Check if we use remote dockerd
    if dc and "localhost" not in dc.api.base_url:
        # dc.api.base_url is most likely tcp://{cmd_opts.remote_docker}
        base_url = dc.api.base_url
        ip_start = base_url.rfind("/") + 1
        ip_end = base_url.rfind(":")
        ip = base_url[ip_start:ip_end]
    else:
        # then dc.api.base_url is most likely http+docker://localhost
        ip = None
    adcm = dw.run_adcm(image=adcm_repo, tag=adcm_tag, remove=False, pull=pull, ip=ip)
    # Create a snapshot from initialized container
    adcm.container.stop()
    with allure.step(f"Commit initialized ADCM container to image {repo}:{tag}"):
        adcm.container.commit(repository=repo, tag=tag)
    adcm.container.remove()
    return {"repo": repo, "tag": tag}


def image_exists(repo, tag, dc=None):
    if dc is None:
        dc = docker.from_env(timeout=120)
    try:
        dc.images.get(name="{}:{}".format(repo, tag))
    except ImageNotFound:
        return False
    return True


def split_tag(image_name: str):
    """
    Split docker image name

    >>> split_tag('fedora/httpd')
    ('fedora/httpd', None)
    >>> split_tag('fedora/httpd:')
    ('fedora/httpd', '')
    >>> split_tag('fedora/httpd:version1.0')
    ('fedora/httpd', 'version1.0')
    """
    image = image_name.split(":")
    if len(image) > 1:
        image_repo = image[0]
        image_tag = image[1]
    else:
        image_repo = image[0]
        image_tag = None
    return image_repo, image_tag


def _wait_for_adcm_container_init(container, container_ip, port, timeout=120):
    if not wait_for_url(f"http://{container_ip}:{port}/api/v1/", timeout):
        additional_message = ""
        try:
            container.kill()
        except APIError:
            additional_message = (
                " \nWARNING: Failed to kill docker container. Try to remove it by hand"
            )
        raise TimeoutError(
            f"ADCM API has not responded in {timeout} seconds{additional_message}"
        )


class ADCM:
    """
    Class that wraps ADCM Api operation over self.api (ADCMApiWrapper)
    and wraps docker over self.container (see docker module for info)
    """

    __slots__ = ("container", "ip", "port", "url", "api")

    def __init__(self, container, ip, port):
        self.container = container
        self.ip = ip
        self.port = port
        self.url = "http://{}:{}".format(self.ip, self.port)
        self.api = ADCMApiWrapper(self.url)

    def stop(self):
        """Stops container"""
        self.container.stop()


class DockerWrapper:
    """Class for connection to local docker daemon and spawn ADCM instances."""

    __slots__ = ("client",)

    def __init__(self, base_url="unix://var/run/docker.sock", dc=None):
        self.client = dc if dc else docker.DockerClient(base_url=base_url, timeout=120)

    # pylint: disable=R0913
    def run_adcm(
        self,
        image=None,
        labels=None,
        remove=True,
        pull=True,
        name=None,
        tag=None,
        ip=None,
        volumes=None,
    ):
        """
        Run ADCM in docker image.
        Return ADCM instance.

        Example:
        adcm = docker.run(image='arenadata/adcm', tag=None, ip='127.0.0.1')

        If tag is None or is not present than a tag 'latest' will be used
        """
        if image is None:
            image = "arenadata/adcm"
        if tag is None:
            tag = "latest"
        if pull:
            self.client.images.pull(image, tag)
        if os.environ.get("BUILD_TAG"):
            if not labels:
                labels = {}
            labels.update({"jenkins-job": os.environ["BUILD_TAG"]})
        if not ip:
            ip = DEFAULT_IP

        container, port = self.adcm_container(
            image=image,
            labels=labels,
            remove=remove,
            name=name,
            tag=tag,
            ip=ip,
            volumes=volumes,
        )

        # If test runner is running in docker then 127.0.0.1
        # will be local container loop interface instead of host loop interface,
        # so we need to establish ADCM API connection using internal docker network
        if ip == DEFAULT_IP and is_docker():
            container_ip = self.client.api.inspect_container(container.id)[
                "NetworkSettings"
            ]["IPAddress"]
            port = "8000"
        else:
            container_ip = ip
        _wait_for_adcm_container_init(container, container_ip, port)
        return ADCM(container, container_ip, port)

    # pylint: disable=R0913
    @allure.step("Run ADCM container from the image {image}:{tag}")
    def adcm_container(
        self,
        image=None,
        labels=None,
        remove=True,
        name=None,
        tag=None,
        ip=None,
        volumes=None,
    ):
        """
        Run ADCM in docker image.
        Return ADCM container and bind port.
        """
        for _ in range(0, CONTAINER_START_RETRY_COUNT):
            port = _find_random_port(ip)
            try:
                container = self.client.containers.run(
                    "{}:{}".format(image, tag),
                    ports={"8000": (ip, port)},
                    volumes=volumes,
                    remove=remove,
                    detach=True,
                    labels=labels,
                    name=name,
                )
                break
            except APIError as err:
                if (
                    "failed: port is already allocated" in err.explanation
                    or "bind: address already in use" in err.explanation  # noqa: W503
                ):
                    # such error excepting leaves created container and there is
                    # no way to clean it other than from docker library
                    pass
                else:
                    raise err
        else:
            raise RetryCountExceeded(
                f"Unable to start container after {CONTAINER_START_RETRY_COUNT} retries"
            )
        with allure.step(f"ADCM API started on {ip}:{port}/api/v1"):
            return container, port



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import suppress
from functools import wraps

import allure
from _pytest.outcomes import Skipped


def assume_step(title, exception=None):
    """
    Allows you to suppress exception within the Allure steps.
    Depending on the type of allowed exception, step will receive different statuses -
    - Skipped if suppress Skipped (from pytest.skip())
    - Failed if suppress AssertionError
    - Broken if other

    :param title: same as allure.step() title
    :param exception: allowed exception

    Examples:

    with assume_step('Skipped step'):
        pytest.skip()
    with assume_step('Failed step', exception=AssertionError):
        raise AssertionError("This assertion don't fail test")
    with assume_step('Broken step', exception=ValueError):
        raise ValueError("This is expected exception")

    """
    if callable(title):
        return AssumeStepContext(title.__name__, exception)(title)
    else:
        return AssumeStepContext(title, exception)


class AssumeStepContext:
    def __init__(self, title, exception=None):
        self.title = title
        self.exceptions = (Skipped, exception) if exception else Skipped
        self.allure_cm = allure.step(title)
        self.suppress = suppress(self.exceptions)

    def __call__(self, f):
        @wraps(f)
        def decorator(*args, **kwargs):
            with self.suppress:
                return allure.step(self.title)(f)(*args, **kwargs)

        return decorator

    def __enter__(self):
        return self.allure_cm.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.allure_cm.__exit__(exc_type, exc_val, exc_tb)
        return self.suppress.__exit__(exc_type, exc_val, exc_tb)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

setup(
    name="adcm_pytest_plugin",
    description="The pytest plugin which includes a set of common tools for ADCM testing",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    version="3.3.8",
    # the following makes a plugin available to pytest
    entry_points={"pytest11": ["adcm_pytest_plugin = adcm_pytest_plugin.plugin"]},
    # custom PyPI classifier for pytest plugins
    install_requires=[
        "pytest",
        "docker",
        "adcm_client>=2020.02.11.12",
        "allure-pytest",
        "requests",
        "version_utils",
        "ifaddr",
        "retry",
        "deprecated",
        "coreapi",
    ],
    classifiers=["Framework :: Pytest"],
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=W0611, W0401, W0614
from argparse import Namespace

import pytest
import requests
from _pytest.config import Config
from version_utils import rpm

from .fixtures import *  # noqa: F401, F403
from .docker_utils import split_tag

options: Namespace = Namespace()


def pytest_configure(config: Config):
    """Set global options to plugin property to access to all pytest options outside of fixtures

    Example:
        from adcm_pytest_plugin.plugin import options
        def some_func():
            adcm_image = options.adcm_image
    """
    global options  # pylint: disable=global-statement
    options.__dict__.update(config.option.__dict__)


# pylint: disable=W0621
def pytest_addoption(parser):

    parser.addoption(
        "--staticimage",
        action="store",
        default=None,
        help="Use pre-initialised ADCM image. "
        "If image doesn't exist then it will be created ones "
        "and can be reused in future test runs. "
        "Useful for tests development. You will save about 40 second on ADCM initialisation "
        "Ex: arenadata/adcm:test or some_repo/some_image:some_tag",
    )

    parser.addoption(
        "--dontstop",
        action="store_true",
        default=False,
        help="Keep ADCM containers running after tests are finished",
    )

    parser.addoption(
        "--adcm-image",
        action="store",
        default=None,
        help="ADCM image to use as base image in tests. "
        "Ex: arenadata/adcm:2021031007, or arenadata/adcm then tag will be 'latest'",
    )

    parser.addoption(
        "--adcm-min-version",
        action="store",
        default=None,
        help="In this mode test will run on all images from arenadata/adcm"
        " that are newer that min version. "
        "Argument format 2020.01.30.15-c4c8b2a or 2020.01.30.15",
    )

    parser.addoption(
        "--adcm-images", nargs="+", help="List of images where to run tests"
    )

    parser.addoption(
        "--nopull", action="store_true", default=False, help="Don't pull image"
    )

    parser.addoption(
        "--remote-executor-host",
        action="store",
        default=None,
        help="This option will be ignored if remote-docker option is passed. "
        "This option is used to initialise ADCM API with external IP "
        "to allow incoming connections from any remote executor (ex. Selenoid). "
        "Test will fail if remote host is unreachable",
    )

    parser.addoption(
        "--remote-docker",
        action="store",
        default=None,
        help="This option will ignore remote-executor-host option "
        "and attach ADCM Api to docker host by default. "
        "This option is used to start ADCM instances on the remote host. "
        "Docker daemon should be running and be available with provided ip:port "
        "Ex: '192.168.1.1:2375'",
    )


def pytest_generate_tests(metafunc):
    """
    If any of launch options adcm-min-version or adcm-images is used
    image fixture undergo parametrization with a list of images to be used
    according to used option
    """
    adcm_min_version = metafunc.config.getoption("adcm_min_version")
    adcm_images = metafunc.config.getoption("adcm_images")

    params, ids = parametrized_by_adcm_version(
        adcm_min_version=adcm_min_version, adcm_images=adcm_images
    )
    if params:
        metafunc.parametrize("image", params, indirect=True, ids=ids)


def parametrized_by_adcm_version(adcm_min_version=None, adcm_images=None):
    params = None
    ids = None
    if adcm_min_version:
        repo = "arenadata/adcm"
        params = [[repo, tag] for tag in _get_adcm_new_versions_tags(adcm_min_version)]
        ids = list(map(lambda x: x[1] if x[1] is not None else "latest", params))
    elif adcm_images:
        params = [[repo, tag] for repo, tag in map(split_tag, adcm_images)]
        ids = adcm_images
    return params, ids


def _get_adcm_new_versions_tags(min_ver):
    response = requests.get(
        "https://hub.docker.com/v2/repositories/arenadata/adcm/tags/?page_size=100"
    )
    for data in response.json()["results"]:
        tag = data.get("name")  # get tag
        if tag.isdigit():  # filter latest tag
            # convert to version format
            version = "%s.%s.%s.%s" % (tag[:4], tag[4:6], tag[6:8], tag[8:10])
            # filter older versions
            if rpm.compare_versions(version, min_ver[:13]) != -1:
                yield version.replace(".", "")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    There is no default info about test stages execution available in pytest
    This hook is meant to store such info im metadata
    """
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)
